import os
import unittest
from unittest.mock import MagicMock

from authlib.jose.errors import BadSignatureError
from flask import Flask
from flask_appbuilder import AppBuilder
from flask_appbuilder.const import AUTH_OAUTH
from flask_appbuilder.exceptions import InvalidLoginAttempt
from flask_appbuilder.exceptions import OAuthProviderUnknown
from flask_appbuilder.utils.legacy import get_sqla_class
import jinja2
import jwt
from tests.const import USERNAME_ADMIN
from tests.const import USERNAME_READONLY
from tests.fixtures.users import create_default_users


class OAuthRegistrationRoleTestCase(unittest.TestCase):
    def setUp(self):
        # start Flask
        self.app = Flask(__name__)
        self.app.jinja_env.undefined = jinja2.StrictUndefined
        self.app.config["SQLALCHEMY_DATABASE_URI"] = os.environ.get(
            "SQLALCHEMY_DATABASE_URI", "sqlite:///"
        )
        self.app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
        self.app.config["AUTH_TYPE"] = AUTH_OAUTH
        self.app.config["OAUTH_PROVIDERS"] = [
            {
                "name": "azure",
                "icon": "fa-windows",
                "token_key": "access_token",
                "remote_app": {
                    "client_id": "CLIENT_ID",
                    "client_secret": "SECRET",
                    "api_base_url": "https://login.microsoftonline.com/TENANT_ID/oauth2",
                    "client_kwargs": {
                        "scope": "User.Read name email profile",
                        "resource": "AZURE_APPLICATION_ID",
                    },
                    "request_token_url": None,
                    "access_token_url": "https://login.microsoftonline.com/"
                    "AZURE_APPLICATION_ID/"
                    "oauth2/token",
                    "authorize_url": "https://login.microsoftonline.com/"
                    "AZURE_APPLICATION_ID/"
                    "oauth2/authorize",
                },
            },
            {
                "name": "auth0",
                "icon": "fa-shield-halved",
                "token_key": "access_token",
                "remote_app": {
                    "client_id": "AUTH0_KEY",
                    "client_secret": "AUTH0_SECRET",
                    "api_base_url": "https://AUTH0_DOMAIN/oauth2/v1/",
                    "client_kwargs": {"scope": "openid profile email groups"},
                    "access_token_url": "https://AUTH0_DOMAIN/oauth/token",
                    "authorize_url": "https://AUTH0_DOMAIN/authorize",
                    "server_metadata_url": "https://AUTH0_DOMAIN/.well-known/"
                    "openid-configuration",
                },
            },
        ]

    def tearDown(self):
        # Remove test user
        with self.app.app_context():
            user_alice = self.appbuilder.sm.find_user("alice")
            if user_alice:
                self.appbuilder.session.delete(user_alice)
                self.appbuilder.session.commit()

        # stop Flask
        self.app = None
        # stop Flask-AppBuilder
        self.appbuilder = None

    def assertOnlyDefaultUsers(self):
        users = self.appbuilder.sm.get_all_users()
        user_names = sorted([user.username for user in users])
        self.assertEqual(user_names, [USERNAME_READONLY, USERNAME_ADMIN])

    # ----------------
    # Userinfo Objects
    # ----------------
    userinfo_alice = {
        "username": "alice",
        "first_name": "Alice",
        "last_name": "Doe",
        "email": "alice@example.com",
        "role_keys": ["GROUP_1", "GROUP_2"],
    }

    # ----------------
    # Unit Tests
    # ----------------
    def test__inactive_user(self):
        """
        OAUTH: test login flow for - inactive user
        """
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            users = self.appbuilder.session.query(sm.user_model).all()
            for user in users:
                self.appbuilder.session.delete(user)
            self.appbuilder.session.commit()
            create_default_users(self.appbuilder.session)
            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.com",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # set user inactive
            new_user.active = False

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was not allowed to log in
            self.assertIsNone(user)

    def test__missing_username(self):
        """
        OAUTH: test login flow for - missing credentials
        """
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # create userinfo with missing info
            userinfo_missing = self.userinfo_alice.copy()
            userinfo_missing["username"] = ""

            # attempt login
            user = sm.auth_user_oauth(userinfo_missing)

            # validate - login failure (missing username)
            self.assertIsNone(user)

            # validate - no users were created
            self.assertOnlyDefaultUsers()

    def test__unregistered(self):
        """
        OAUTH: test login flow for - unregistered user
        """
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the AUTH_USER_REGISTRATION_ROLE role
            self.assertEqual(user.roles, [sm.find_role("Public")])

            # validate - user was given the correct attributes
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.com")

    def test__unregistered__no_self_register(self):
        """
        OAUTH: test login flow for - unregistered user - no self-registration
        """
        self.app.config["AUTH_USER_REGISTRATION"] = False
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was not allowed to log in
            self.assertIsNone(user)

            # validate - no users were registered
            self.assertOnlyDefaultUsers()

    def test__unregistered__single_role(self):
        """
        OAUTH: test login flow for - unregistered user
                                   - single role mapping
        """
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "GROUP_1": ["Admin"],
            "GROUP_2": ["User"],
        }
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertIn(sm.find_role("Admin"), user.roles)
            self.assertIn(sm.find_role("User"), user.roles)
            self.assertIn(sm.find_role("Public"), user.roles)

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.com")

    def test__unregistered__multi_role(self):
        """
        OAUTH: test login flow for - unregistered user - multi role mapping
        """
        self.app.config["AUTH_ROLES_MAPPING"] = {"GROUP_1": ["Admin", "User"]}
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertIn(sm.find_role("Admin"), user.roles)
            self.assertIn(sm.find_role("Public"), user.roles)
            self.assertIn(sm.find_role("User"), user.roles)

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.com")

    def test__unregistered__jmespath_role(self):
        """
        OAUTH: test login flow for - unregistered user - jmespath registration role
        """
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config[
            "AUTH_USER_REGISTRATION_ROLE_JMESPATH"
        ] = "contains(['alice'], username) && 'User' || 'Public'"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertListEqual(user.roles, [sm.find_role("User")])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.com")

    def test__registered__multi_role__no_role_sync(self):
        """
        OAUTH: test login flow for - registered user - multi role mapping - no login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_MAPPING"] = {"GROUP_1": ["Admin", "User"]}
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = False
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.com",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given no roles
            self.assertListEqual(user.roles, [])

    def test__registered__multi_role__with_role_sync(self):
        """
        OAUTH: test login flow for - registered user - multi role mapping - with login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_MAPPING"] = {"GROUP_1": ["Admin", "User"]}
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = True
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.com",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given the correct roles
            self.assertSetEqual(
                set(user.roles), {sm.find_role("Admin"), sm.find_role("User")}
            )

    def test__registered__jmespath_role__no_role_sync(self):
        """
        OAUTH: test login flow for - registered user - jmespath registration role - no login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = False
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config[
            "AUTH_USER_REGISTRATION_ROLE_JMESPATH"
        ] = "contains(['alice'], username) && 'User' || 'Public'"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.com",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given no roles
            self.assertListEqual(user.roles, [])

    def test__registered__jmespath_role__with_role_sync(self):
        """
        OAUTH: test login flow for - registered user - jmespath registration role - with login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = True
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config[
            "AUTH_USER_REGISTRATION_ROLE_JMESPATH"
        ] = "contains(['alice'], username) && 'User' || 'Public'"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.com",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_oauth(self.userinfo_alice)

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given the correct roles
            self.assertListEqual(user.roles, [sm.find_role("User")])

    def test_oauth_user_info_getter(self):
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)

            @self.appbuilder.sm.oauth_user_info_getter
            def user_info_getter(sm, provider, response):
                return {"username": "test"}

            self.assertEqual(self.appbuilder.sm.oauth_user_info, user_info_getter)
            self.assertEqual(
                self.appbuilder.sm.oauth_user_info("azure", {"claim": 1}),
                {"username": "test"},
            )

    def test_oauth_user_info_unknown_provider(self):
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            with self.assertRaises(OAuthProviderUnknown):
                self.appbuilder.sm.oauth_user_info("unknown", {})

    def test_oauth_user_info_azure_email_upn(self):
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            claims = {
                "aud": "test-aud",
                "iss": "https://sts.windows.net/test/",
                "iat": 7282182129,
                "nbf": 7282182129,
                "exp": 1000000000,
                "amr": ["pwd"],
                "email": "test@gmail.com",
                "upn": "test@upn.com",
                "family_name": "user",
                "given_name": "test",
                "idp": "live.com",
                "name": "Test user",
                "oid": "b1a54a40-8dfa-4a6d-a2b8-f90b84d4b1df",
                "unique_name": "live.com#test@gmail.com",
                "ver": "1.0",
            }

            # Create an unsigned JWT
            unsigned_jwt = jwt.encode(claims, key=None, algorithm="none")
            user_info = self.appbuilder.sm.get_oauth_user_info(
                "azure", {"access_token": "", "id_token": unsigned_jwt}
            )
            self.assertEqual(
                user_info,
                {
                    "email": "test@upn.com",
                    "first_name": "test",
                    "last_name": "user",
                    "role_keys": [],
                    "username": "b1a54a40-8dfa-4a6d-a2b8-f90b84d4b1df",
                },
            )

    def test_oauth_user_info_azure(self):
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            claims = {
                "aud": "test-aud",
                "iss": "https://sts.windows.net/test/",
                "iat": 7282182129,
                "nbf": 7282182129,
                "exp": 1000000000,
                "amr": ["pwd"],
                "email": "test@gmail.com",
                "family_name": "user",
                "given_name": "test",
                "idp": "live.com",
                "name": "Test user",
                "oid": "b1a54a40-8dfa-4a6d-a2b8-f90b84d4b1df",
                "unique_name": "live.com#test@gmail.com",
                "ver": "1.0",
            }

            # Create an unsigned JWT
            unsigned_jwt = jwt.encode(claims, key=None, algorithm="none")
            user_info = self.appbuilder.sm.get_oauth_user_info(
                "azure", {"access_token": "", "id_token": unsigned_jwt}
            )
            self.assertEqual(
                user_info,
                {
                    "email": "test@gmail.com",
                    "first_name": "test",
                    "last_name": "user",
                    "role_keys": [],
                    "username": "b1a54a40-8dfa-4a6d-a2b8-f90b84d4b1df",
                },
            )

    def test_oauth_user_info_azure_with_jwt_validation(self):
        self.app.config["OAUTH_PROVIDERS"] = [
            {
                "name": "azure",
                "icon": "fa-windows",
                "token_key": "access_token",
                "remote_app": {
                    "client_id": "CLIENT_ID",
                    "client_secret": "SECRET",
                    "api_base_url": "https://login.microsoftonline.com/TENANT_ID/oauth2",
                    "client_kwargs": {
                        "scope": "User.Read name email profile",
                        "resource": "AZURE_APPLICATION_ID",
                        "verify_signature": True,
                    },
                    "request_token_url": None,
                    "access_token_url": "https://login.microsoftonline.com/"
                    "AZURE_APPLICATION_ID/"
                    "oauth2/token",
                    "authorize_url": "https://login.microsoftonline.com/"
                    "AZURE_APPLICATION_ID/"
                    "oauth2/authorize",
                },
            }
        ]
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            claims = {
                "aud": "test-aud",
                "iss": "https://sts.windows.net/test/",
                "iat": 1696601585,
                "nbf": 1696601585,
                "exp": 7282182129,  # 100 years from now ;)
                "amr": ["pwd"],
                "email": "test@gmail.com",
                "family_name": "user",
                "given_name": "test",
                "idp": "live.com",
                "name": "Test user",
                "oid": "b1a54a40-8dfa-4a6d-a2b8-f90b84d4b1df",
                "unique_name": "live.com#test@gmail.com",
                "ver": "1.0",
            }
            from unittest.mock import MagicMock

            private_key = """-----BEGIN PRIVATE KEY-----
MIICdgIBADANBgkqhkiG9w0BAQEFAASCAmAwggJcAgEAAoGBALeDojEka93XZ/J8
bDGgn2MIHykafgCx2D6wTZgmmhzpRH7/k7J/WSsqG6eSFg38mGJukPCa4dcG8dCL
meajEf2g4IoaYiE55yXs0ou/tixBJI8wRY+NfCluxgIcHdKhZISVO6CkR5r7diN/
SLHPsFnDd0UiMJ5c48UsJwk8T5T7AgMBAAECgYEAqalrVB+mEi1KDud1Z9RmRzqF
BI1XnPDPSfXZZyeZJ82J5BgJxubx23RMqPnopfm4MJikK64lyZTED9hg6tgskk1X
J9pc7iyU4PQf+tx4tvElyOL4OSqGss/tQHtHz76hNOR1kxeCcJsJG+WS8P0/Kmj1
0IoYKLFlb5AHr6KqDGECQQDZ0qKIzxdmZj3gSsNldc4oOQOKJgd1QSDGCOqR9p7f
oj7nuOPRVgnztqXALhNhpZXYJq8dWmpGYFi+EC1piRUDAkEA162gPgGzUJAIyhUM
sA6Uy9v64nqBnlygVpofhdvyznSf/KUsmWQZv7gpMMXnIGAQP+rqM1gJvuRtodml
hUeSqQJAHJH4J6GiHBhE/WpQ/rnY9IWl5TTfvY1xUwhQXBzQ8dxCC/rARvDWFVVb
oD1q5V/mq5dHWL5HOjvg5+0PR8xnKQJAMOdBik3AZugB1jBnrBPiUUcT3/5/HXVL
NdfEhgmVSJLRI+wf7LfxzrLnRBPbkE+334ZYjEPOEeahpS1AhrPv4QJAHpap1I+v
8m+N5G/MppasppHLJmXhnFeQsnBX7XcdYiCqHikuBlIzoQ0Cj5xbkfgMMCVORO64
r9+EFRsxA5GNYA==
-----END PRIVATE KEY-----"""
            # Create an unsigned JWT
            unsigned_jwt = jwt.encode(
                claims, key=private_key, algorithm="RS256", headers={"kid": "1"}
            )
            self.appbuilder.sm._get_microsoft_jwks = MagicMock(
                return_value={
                    "keys": [
                        {
                            "alg": "RS256",
                            "e": "AQAB",
                            "kid": "1",
                            "kty": "RSA",
                            "n": "t4OiMSRr3ddn8nxsMaCfYwgfKRp-ALHYPrBNmCaaHOlEfv-"
                            "Tsn9ZKyobp5IWDfyYYm6Q8Jrh1wbx0IuZ5qMR_aDgihpiITnnJezSi7-"
                            "2LEEkjzBFj418KW7GAhwd0qFkhJU7oKRHmvt2I39Isc-wWcN3RSIwnlz"
                            "jxSwnCTxPlPs",
                            "use": "sig",
                            "x5c": [
                                "MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQC3g6IxJGvd12fyfGwx"
                                "oJ9jCB8pGn4Asdg+sE2YJpoc6UR+/5Oyf1krKhunkhYN/JhibpDwmuHX"
                                "BvHQi5nmoxH9oOCKGmIhOecl7NKLv7YsQSSPMEWPjXwpbsYCHB3SoWSE"
                                "lTugpEea+3Yjf0ixz7BZw3dFIjCeXOPFLCcJPE+U+wIDAQAB"
                            ],
                        }
                    ]
                }
            )
            user_info = self.appbuilder.sm.get_oauth_user_info(
                "azure", {"access_token": "", "id_token": unsigned_jwt}
            )
            self.assertEqual(
                user_info,
                {
                    "email": "test@gmail.com",
                    "first_name": "test",
                    "last_name": "user",
                    "role_keys": [],
                    "username": "b1a54a40-8dfa-4a6d-a2b8-f90b84d4b1df",
                },
            )

    def test_oauth_user_info_auth0(self):
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)

            self.appbuilder.sm.oauth_remotes["auth0"].userinfo = MagicMock(
                return_value={
                    "email": "test@gmail.com",
                    "given_name": "test",
                    "family_name": "user",
                    "role_keys": [],
                    "sub": "test-sub",
                }
            )

            user_info = self.appbuilder.sm.get_oauth_user_info(
                "auth0", {"access_token": "", "id_token": ""}
            )
            self.assertEqual(
                user_info,
                {
                    "email": "test@gmail.com",
                    "first_name": "test",
                    "last_name": "user",
                    "role_keys": [],
                    "username": "auth0_test-sub",
                },
            )

    def test_reset_password_view_not_registered_with_oauth(self):
        """
        OAUTH: test that ResetMyPasswordView is not registered when using OAuth
        authentication
        """
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)

            for view in self.appbuilder.baseviews:
                if view.__class__.__name__ == "ResetMyPasswordView":
                    self.fail(
                        "ResetMyPasswordView should not be registered when using OAuth "
                        "authentication"
                    )

            # Also verify that the view is not accessible via URL
            with self.app.test_client() as client:
                response = client.get("/resetmypassword/form")
                # Should return 404 since the view is not registered
                self.assertEqual(response.status_code, 404)


class OAuthAuthentikTestCase(unittest.TestCase):
    def setUp(self):
        # start Flask
        self.app = Flask(__name__)
        self.app.jinja_env.undefined = jinja2.StrictUndefined
        self.app.config["SQLALCHEMY_DATABASE_URI"] = os.environ.get(
            "SQLALCHEMY_DATABASE_URI", "sqlite:///"
        )
        self.app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
        self.app.config["AUTH_TYPE"] = AUTH_OAUTH
        self.app.config["OAUTH_PROVIDERS"] = [
            {
                "name": "authentik",
                "token_key": "access_token",
                "icon": "fa-fingerprint",
                "remote_app": {
                    "api_base_url": "https://authentik.mydomain.com",
                    "client_kwargs": {
                        "scope": "email profile",
                        "verify_signature": False,
                    },
                    "access_token_url": (
                        "https://authentik.mydomain.com" "/application/o/token/"
                    ),
                    "authorize_url": (
                        "https://authentik.mydomain.com/" "application/o/authorize/"
                    ),
                    "request_token_url": None,
                    "client_id": "CLIENT_ID",
                    "client_secret": "CLIENT_SECRET",
                },
            },
        ]

    def tearDown(self):
        # Remove test user
        with self.app.app_context():
            user_alice = self.appbuilder.sm.find_user("alice")
            if user_alice:
                self.appbuilder.session.delete(user_alice)
                self.appbuilder.session.commit()

    # ----------------
    # Unit Tests
    # ----------------
    def test_oauth_user_info_authentik(self):
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            claims = {
                "iss": "https://authentik.mydomain.com/application/o/"
                "flask-appbuilder-test/",
                "sub": "2ac1102e7cf5a4b1cb2dd5adbe4761c551691ecd88991f78d0195d4d3d0cfcfa",
                "aud": "CLIENT_ID",
                "exp": 1703257941,
                "iat": 1700665941,
                "auth_time": 7282182129,  # 100 years from now ;)
                "acr": "goauthentik.io/providers/oauth2/default",
                "at_hash": "cAydO2DJMi_ZL6opx3eUdw",
                "email": "alice@example.com",
                "email_verified": True,
                "name": "Alice",
                "given_name": "Alice Doe",
                "preferred_username": "alice@example.com",
                "nickname": "alice",
                "groups": ["GROUP_1", "GROUP_2"],
            }

            # Create an unsigned JWT
            unsigned_jwt = jwt.encode(claims, key=None, algorithm="none")
            user_info = self.appbuilder.sm.get_oauth_user_info(
                "authentik", {"access_token": "", "id_token": unsigned_jwt}
            )
            self.assertEqual(
                user_info,
                {
                    "email": "alice@example.com",
                    "first_name": "Alice Doe",
                    "role_keys": ["GROUP_1", "GROUP_2"],
                    "username": "alice",
                },
            )

    def test_oauth_user_info_authentik_with_jwt_validation(self):
        self.app.config["OAUTH_PROVIDERS"] = [
            {
                "name": "authentik",
                "token_key": "access_token",
                "icon": "fa-fingerprint",
                "remote_app": {
                    "api_base_url": "https://authentik.mydomain.com",
                    "client_kwargs": {
                        "scope": "email profile",
                        "verify_signature": True,
                    },
                    "access_token_url": (
                        "https://authentik.mydomain.com" "/application/o/token/"
                    ),
                    "authorize_url": (
                        "https://authentik.mydomain.com/" "application/o/authorize/"
                    ),
                    "request_token_url": None,
                    "client_id": "CLIENT_ID",
                    "client_secret": "CLIENT_SECRET",
                    "jwks_uri": (
                        "https://authentik.mydomain.com/"
                        "application/o/APPLICATION_NAME/jwks/"
                    ),
                },
            },
        ]
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            claims = {
                "iss": "https://authentik.mydomain.com/application/o/"
                "flask-appbuilder-test/",
                "sub": "2ac1102e7cf5a4b1cb2dd5adbe4761c551691ecd88991f78d0195d4d3d0cfcfa",
                "aud": "CLIENT_ID",
                "exp": 3203257941,
                "iat": 1700665941,
                "auth_time": 7282182129,  # 100 years from now ;)
                "acr": "goauthentik.io/providers/oauth2/default",
                "at_hash": "cAydO2DJMi_ZL6opx3eUdw",
                "email": "alice@example.com",
                "email_verified": True,
                "name": "Alice",
                "given_name": "Alice Doe",
                "preferred_username": "alice@example.com",
                "nickname": "alice",
                "groups": ["GROUP_1", "GROUP_2"],
            }
            from unittest.mock import MagicMock

            private_key = """-----BEGIN PRIVATE KEY-----
MIICdgIBADANBgkqhkiG9w0BAQEFAASCAmAwggJcAgEAAoGBALeDojEka93XZ/J8
bDGgn2MIHykafgCx2D6wTZgmmhzpRH7/k7J/WSsqG6eSFg38mGJukPCa4dcG8dCL
meajEf2g4IoaYiE55yXs0ou/tixBJI8wRY+NfCluxgIcHdKhZISVO6CkR5r7diN/
SLHPsFnDd0UiMJ5c48UsJwk8T5T7AgMBAAECgYEAqalrVB+mEi1KDud1Z9RmRzqF
BI1XnPDPSfXZZyeZJ82J5BgJxubx23RMqPnopfm4MJikK64lyZTED9hg6tgskk1X
J9pc7iyU4PQf+tx4tvElyOL4OSqGss/tQHtHz76hNOR1kxeCcJsJG+WS8P0/Kmj1
0IoYKLFlb5AHr6KqDGECQQDZ0qKIzxdmZj3gSsNldc4oOQOKJgd1QSDGCOqR9p7f
oj7nuOPRVgnztqXALhNhpZXYJq8dWmpGYFi+EC1piRUDAkEA162gPgGzUJAIyhUM
sA6Uy9v64nqBnlygVpofhdvyznSf/KUsmWQZv7gpMMXnIGAQP+rqM1gJvuRtodml
hUeSqQJAHJH4J6GiHBhE/WpQ/rnY9IWl5TTfvY1xUwhQXBzQ8dxCC/rARvDWFVVb
oD1q5V/mq5dHWL5HOjvg5+0PR8xnKQJAMOdBik3AZugB1jBnrBPiUUcT3/5/HXVL
NdfEhgmVSJLRI+wf7LfxzrLnRBPbkE+334ZYjEPOEeahpS1AhrPv4QJAHpap1I+v
8m+N5G/MppasppHLJmXhnFeQsnBX7XcdYiCqHikuBlIzoQ0Cj5xbkfgMMCVORO64
r9+EFRsxA5GNYA==
-----END PRIVATE KEY-----"""
            # Create a signed JWT
            signed_jwt = jwt.encode(
                claims, key=private_key, algorithm="RS256", headers={"kid": "1"}
            )
            self.appbuilder.sm._get_authentik_jwks = MagicMock(
                return_value={
                    "keys": [
                        {
                            "alg": "RS256",
                            "e": "AQAB",
                            "kid": "1",
                            "kty": "RSA",
                            "n": "t4OiMSRr3ddn8nxsMaCfYwgfKRp-ALHYPrBNmCaaHOlEfv-"
                            "Tsn9ZKyobp5IWDfyYYm6Q8Jrh1wbx0IuZ5qMR_aDgihpiITnnJezSi7-"
                            "2LEEkjzBFj418KW7GAhwd0qFkhJU7oKRHmvt2I39Isc-wWcN3RSIwnlz"
                            "jxSwnCTxPlPs",
                            "use": "sig",
                            "x5c": [
                                "MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQC3g6IxJGvd12fyfGwx"
                                "oJ9jCB8pGn4Asdg+sE2YJpoc6UR+/5Oyf1krKhunkhYN/JhibpDwmuHX"
                                "BvHQi5nmoxH9oOCKGmIhOecl7NKLv7YsQSSPMEWPjXwpbsYCHB3SoWSE"
                                "lTugpEea+3Yjf0ixz7BZw3dFIjCeXOPFLCcJPE+U+wIDAQAB"
                            ],
                        }
                    ]
                }
            )
            user_info = self.appbuilder.sm.get_oauth_user_info(
                "authentik", {"access_token": "", "id_token": signed_jwt}
            )
            self.assertEqual(
                user_info,
                {
                    "email": "alice@example.com",
                    "first_name": "Alice Doe",
                    "role_keys": ["GROUP_1", "GROUP_2"],
                    "username": "alice",
                },
            )

    def test_oauth_user_info_authentik_with_jwt_validation_wrong_signature(self):
        """
        Tests if the get_user_info raises an exception
        if the token is signed by a different JKWS
        """
        self.app.config["OAUTH_PROVIDERS"] = [
            {
                "name": "authentik",
                "token_key": "access_token",
                "icon": "fa-fingerprint",
                "remote_app": {
                    "api_base_url": "https://authentik.mydomain.com",
                    "client_kwargs": {
                        "scope": "email profile",
                        "verify_signature": True,
                    },
                    "access_token_url": (
                        "https://authentik.mydomain.com" "/application/o/token/"
                    ),
                    "authorize_url": (
                        "https://authentik.mydomain.com/" "application/o/authorize/"
                    ),
                    "request_token_url": None,
                    "client_id": "CLIENT_ID",
                    "client_secret": "CLIENT_SECRET",
                    "jwks_uri": (
                        "https://authentik.mydomain.com/"
                        "application/o/APPLICATION_NAME/jwks/"
                    ),
                },
            },
        ]
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            claims = {
                "iss": "https://authentik.mydomain.com/application/o/"
                "flask-appbuilder-test/",
                "sub": "2ac1102e7cf5a4b1cb2dd5adbe4761c551691ecd88991f78d0195d4d3d0cfcfa",
                "aud": "CLIENT_ID",
                "exp": 1703257941,
                "iat": 1700665941,
                "auth_time": 7282182129,  # 100 years from now ;)
                "acr": "goauthentik.io/providers/oauth2/default",
                "at_hash": "cAydO2DJMi_ZL6opx3eUdw",
                "email": "alice@example.com",
                "email_verified": True,
                "name": "Alice",
                "given_name": "Alice Doe",
                "preferred_username": "alice@example.com",
                "nickname": "alice",
                "groups": ["GROUP_1", "GROUP_2"],
            }
            from unittest.mock import MagicMock

            private_key = """-----BEGIN PRIVATE KEY-----
MIIBVAIBADANBgkqhkiG9w0BAQEFAASCAT4wggE6AgEAAkEAqPfgaTEWEP3S9w0t
gsicURfo+nLW09/0KfOPinhYZ4ouzU+3xC4pSlEp8Ut9FgL0AgqNslNaK34Kq+NZ
jO9DAQIDAQABAkAgkuLEHLaqkWhLgNKagSajeobLS3rPT0Agm0f7k55FXVt743hw
Ngkp98bMNrzy9AQ1mJGbQZGrpr4c8ZAx3aRNAiEAoxK/MgGeeLui385KJ7ZOYktj
hLBNAB69fKwTZFsUNh0CIQEJQRpFCcydunv2bENcN/oBTRw39E8GNv2pIcNxZkcb
NQIgbYSzn3Py6AasNj6nEtCfB+i1p3F35TK/87DlPSrmAgkCIQDJLhFoj1gbwRbH
/bDRPrtlRUDDx44wHoEhSDRdy77eiQIgE6z/k6I+ChN1LLttwX0galITxmAYrOBh
BVl433tgTTQ=
-----END PRIVATE KEY-----"""
            # Create a signed JWT
            wrong_signed_jwt = jwt.encode(
                claims, key=private_key, algorithm="RS256", headers={"kid": "1"}
            )
            self.appbuilder.sm._get_authentik_jwks = MagicMock(
                return_value={
                    "keys": [
                        {
                            "alg": "RS256",
                            "e": "AQAB",
                            "kid": "1",
                            "kty": "RSA",
                            "n": "t4OiMSRr3ddn8nxsMaCfYwgfKRp-ALHYPrBNmCaaHOlEfv-"
                            "Tsn9ZKyobp5IWDfyYYm6Q8Jrh1wbx0IuZ5qMR_aDgihpiITnnJezSi7-"
                            "2LEEkjzBFj418KW7GAhwd0qFkhJU7oKRHmvt2I39Isc-wWcN3RSIwnlz"
                            "jxSwnCTxPlPs",
                            "use": "sig",
                            "x5c": [
                                "MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQC3g6IxJGvd12fyfGwx"
                                "oJ9jCB8pGn4Asdg+sE2YJpoc6UR+/5Oyf1krKhunkhYN/JhibpDwmuHX"
                                "BvHQi5nmoxH9oOCKGmIhOecl7NKLv7YsQSSPMEWPjXwpbsYCHB3SoWSE"
                                "lTugpEea+3Yjf0ixz7BZw3dFIjCeXOPFLCcJPE+U+wIDAQAB"
                            ],
                        }
                    ]
                }
            )
            with self.assertRaises(BadSignatureError):
                self.appbuilder.sm.get_oauth_user_info(
                    "authentik", {"access_token": "", "id_token": wrong_signed_jwt}
                )

    def test_oauth_user_info_authentik_with_jwt_validation_without_signature(self):
        """
        Tests if an unsigned token raises an error if verify_signature is set to True
        """
        self.app.config["OAUTH_PROVIDERS"] = [
            {
                "name": "authentik",
                "token_key": "access_token",
                "icon": "fa-fingerprint",
                "remote_app": {
                    "api_base_url": "https://authentik.mydomain.com",
                    "client_kwargs": {
                        "scope": "email profile",
                        "verify_signature": True,
                    },
                    "access_token_url": (
                        "https://authentik.mydomain.com" "/application/o/token/"
                    ),
                    "authorize_url": (
                        "https://authentik.mydomain.com/" "application/o/authorize/"
                    ),
                    "request_token_url": None,
                    "client_id": "CLIENT_ID",
                    "client_secret": "CLIENT_SECRET",
                    "jwks_uri": (
                        "https://authentik.mydomain.com/"
                        "application/o/APPLICATION_NAME/jwks/"
                    ),
                },
            },
        ]
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            claims = {
                "iss": "https://authentik.mydomain.com/application/o/"
                "flask-appbuilder-test/",
                "sub": "2ac1102e7cf5a4b1cb2dd5adbe4761c551691ecd88991f78d0195d4d3d0cfcfa",
                "aud": "CLIENT_ID",
                "exp": 1703257941,
                "iat": 1700665941,
                "auth_time": 7282182129,  # 100 years from now ;)
                "acr": "goauthentik.io/providers/oauth2/default",
                "at_hash": "cAydO2DJMi_ZL6opx3eUdw",
                "email": "alice@example.com",
                "email_verified": True,
                "name": "Alice",
                "given_name": "Alice Doe",
                "preferred_username": "alice@example.com",
                "nickname": "alice",
                "groups": ["GROUP_1", "GROUP_2"],
            }
            from unittest.mock import MagicMock

            unsigned_jwt = jwt.encode(claims, key=None, algorithm="none")
            self.appbuilder.sm._get_authentik_jwks = MagicMock(return_value={})
            with self.assertRaises(InvalidLoginAttempt):
                self.appbuilder.sm.get_oauth_user_info(
                    "authentik", {"access_token": "", "id_token": unsigned_jwt}
                )



from unittest.mock import patch

from flask_appbuilder import ModelView
from flask_appbuilder.exceptions import PasswordComplexityValidationError
from flask_appbuilder.models.sqla.filters import FilterEqual
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder.security.sqla.models import Group, User
from tests.base import BaseMVCTestCase
from tests.const import (
    INVALID_LOGIN_STRING,
    PASSWORD_ADMIN,
    PASSWORD_READONLY,
    USERNAME_ADMIN,
    USERNAME_READONLY,
)
from tests.fixtures.data_models import model1_data
from tests.sqla.models import Model1, Model2

PASSWORD_COMPLEXITY_ERROR = (
    "Must have at least two capital letters, "
    "one special character, two digits, three lower case letters and "
    "a minimal length of 10"
)


def custom_password_validator(password: str) -> None:
    """
    A simplistic example for a password validator
    """
    if password != "password":
        raise PasswordComplexityValidationError("Password must be password")


class MVCSecurityTestCase(BaseMVCTestCase):
    def setUp(self):
        super().setUp()
        self.client = self.app.test_client()

        class Model2View(ModelView):
            datamodel = SQLAInterface(Model2)
            list_columns = [
                "field_integer",
                "field_float",
                "field_string",
                "field_method",
                "group.field_string",
            ]
            edit_form_query_rel_fields = {
                "group": [["field_string", FilterEqual, "test1"]]
            }
            add_form_query_rel_fields = {
                "group": [["field_string", FilterEqual, "test0"]]
            }

            order_columns = ["field_string", "group.field_string"]

        self.appbuilder.add_view(Model2View, "Model2")

        class Model1View(ModelView):
            datamodel = SQLAInterface(Model1)
            related_views = [Model2View]
            list_columns = ["field_string", "field_integer"]

        self.appbuilder.add_view(Model1View, "Model1", category="Model1")

    def test_sec_login_no_cache(self):
        """
        Test Security Login, no cache directives
        """
        rv = self.client.get("/login/")
        assert rv.status_code == 200
        assert (
            rv.headers.get("Cache-Control")
            == "no-store, no-cache, must-revalidate, max-age=0"
        )
        assert rv.headers["Pragma"] == "no-cache"
        assert rv.headers["Expires"] == "0"

    def test_sec_login(self):
        """
        Test Security Login, Logout, invalid login, invalid access
        """

        # Try to List and Redirect to Login
        rv = self.client.get("/model1view/list/")
        self.assertEqual(rv.status_code, 302)
        rv = self.client.get("/model2view/list/")
        self.assertEqual(rv.status_code, 302)

        # Login and list with admin
        self.browser_login(self.client, USERNAME_ADMIN, PASSWORD_ADMIN)
        rv = self.client.get("/model1view/list/")
        self.assertEqual(rv.status_code, 200)
        rv = self.client.get("/model2view/list/")
        self.assertEqual(rv.status_code, 200)

        # Logout and and try to list
        self.browser_logout(self.client)
        rv = self.client.get("/model1view/list/")
        self.assertEqual(rv.status_code, 302)
        rv = self.client.get("/model2view/list/")
        self.assertEqual(rv.status_code, 302)

        # Invalid Login
        rv = self.browser_login(self.client, USERNAME_ADMIN, "wrong_password")
        data = rv.data.decode("utf-8")
        self.assertIn(INVALID_LOGIN_STRING, data)

    def test_login_invalid_user(self):
        """
        Test Security Login, Logout, invalid login, invalid access
        """
        self.browser_logout(self.client)
        test_username = "testuser"
        test_password = "password"
        test_user = self.create_user(
            self.appbuilder,
            test_username,
            test_password,
            "Admin",
            "user",
            "user",
            "testuser@fab.org",
        )
        # Login and list with admin
        self.browser_login(self.client, test_username, test_password)
        rv = self.client.get("/model1view/list/")
        self.assertEqual(rv.status_code, 200)

        # Using the same session make sure the user is not allowed to access when
        # the user is deactivated
        test_user.active = False
        self.appbuilder.session.merge(test_user)
        self.appbuilder.session.commit()
        rv = self.client.get("/model1view/list/")
        self.assertEqual(rv.status_code, 302)

        self.appbuilder.session.delete(test_user)
        self.appbuilder.session.commit()

    def test_db_login_no_next_url(self):
        """
        Test Security no next URL
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client, USERNAME_ADMIN, PASSWORD_ADMIN, follow_redirects=False
        )
        assert response.location == "/"

    def test_db_login_valid_next_url(self):
        """
        Test Security valid partial next URL
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="/users/list/",
            follow_redirects=False,
        )
        assert response.location == "/users/list/"

    def test_db_login_valid_http_scheme_url(self):
        """
        Test Security valid http scheme next URL
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="http://localhost/path",
            follow_redirects=False,
        )
        assert response.location == "http://localhost/path"

    def test_db_login_valid_https_scheme_url(self):
        """
        Test Security valid https scheme next URL
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="https://localhost/path",
            follow_redirects=False,
        )
        assert response.location == "https://localhost/path"

    def test_db_login_invalid_external_next_url(self):
        """
        Test Security invalid external next URL
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="https://google.com",
            follow_redirects=False,
        )
        assert response.location == "/"

    def test_db_login_invalid_scheme_next_url(self):
        """
        Test Security invalid scheme next URL
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="ftp://sample",
            follow_redirects=False,
        )
        assert response.location == "/"

    def test_db_login_invalid_localhost_file_next_url(self):
        """
        Test Security invalid path to localhost file next URL
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="file:///path",
            follow_redirects=False,
        )
        assert response.location == "/"

    def test_db_login_invalid_no_netloc_with_scheme_next_url(self):
        """
        Test Security invalid next URL with no netloc but with scheme
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="http:///sample.com ",
            follow_redirects=False,
        )
        assert response.location == "/"

    def test_login_next_url_spoofed_host_header_disallowed(self):
        """
        Ensure a spoofed Host header does not allow redirection to an untrusted domain
        """
        self.app.config["FAB_SAFE_REDIRECT_HOSTS"] = ["localhost"]  # trusted dev host
        self.browser_logout(self.client)

        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="https://example.com/",
            follow_redirects=False,
            headers={"Host": "example.com"},
        )

        assert response.status_code == 302
        assert response.location == "/"

    def test_login_next_url_spoofed_host_header_allowed_config(self):
        """
        Ensure a spoofed Host header does not allow redirection to an untrusted domain
        """
        self.app.config["FAB_SAFE_REDIRECT_HOSTS"] = ["localhost"]  # trusted dev host
        self.browser_logout(self.client)

        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="https://localhost/something",
            follow_redirects=False,
            headers={"Host": "example.com"},
        )

        assert response.status_code == 302
        assert response.location == "https://localhost/something"

    def test_login_next_url_allowed_config_wildcard(self):
        """
        Ensure a spoofed Host header does not allow redirection to an untrusted domain
        """
        self.app.config["FAB_SAFE_REDIRECT_HOSTS"] = ["*.localhost"]  # trusted dev host
        self.browser_logout(self.client)

        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="https://example.localhost/something",
            follow_redirects=False,
            headers={"Host": "example.localhost"},
        )

        assert response.status_code == 302
        assert response.location == "https://example.localhost/something"

    def test_login_next_url_spoofed_host_header_allowed(self):
        """
        Ensure a spoofed Host header does not allow redirection to an untrusted domain
        """
        self.browser_logout(self.client)

        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="https://example.com/",
            follow_redirects=False,
            headers={"Host": "example.com"},
        )

        assert response.status_code == 302
        assert response.location == "https://example.com/"

    def test_db_login_invalid_control_characters_next_url(self):
        """
        Test Security invalid next URL with control characters
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            PASSWORD_ADMIN,
            next_url="\u0001" + "sample.com",
            follow_redirects=False,
        )
        assert response.location == "/"

    def test_db_login_failed_keep_next_url(self):
        """
        Test Security Keeping next url after failed login attempt
        """
        self.browser_logout(self.client)
        response = self.browser_login(
            self.client,
            USERNAME_ADMIN,
            f"wrong_{PASSWORD_ADMIN}",
            next_url="/users/list/",
            follow_redirects=False,
        )
        response = self.client.post(
            response.location,
            data=dict(username=USERNAME_ADMIN, password=PASSWORD_ADMIN),
            follow_redirects=False,
        )

        assert response.location == "/users/list/"

    def test_auth_builtin_roles(self):
        """
        Test Security builtin roles readonly
        """
        client = self.app.test_client()
        self.browser_login(client, USERNAME_READONLY, PASSWORD_READONLY)
        with model1_data(self.appbuilder.session, 1) as model_data:
            model_id = model_data[0].id
            # Test authorized GET
            rv = client.get("/model1view/list/")
            self.assertEqual(rv.status_code, 200)
            # Test authorized SHOW
            rv = client.get(f"/model1view/show/{model_id}")
            self.assertEqual(rv.status_code, 200)
            # Test unauthorized EDIT
            rv = client.get(f"/model1view/edit/{model_id}")
            self.assertEqual(rv.status_code, 302)
            # Test unauthorized DELETE
            rv = client.get(f"/model1view/delete/{model_id}")
            self.assertEqual(rv.status_code, 302)

    def test_sec_reset_password(self):
        """
        Test Security reset password
        """
        client = self.app.test_client()
        admin_user = self.appbuilder.sm.find_user(username=USERNAME_ADMIN)
        # Try Reset My password
        rv = client.get(
            f"/users/action/resetmypassword/{admin_user.id}", follow_redirects=True
        )
        # Werkzeug update to 0.15.X sends this action to wrong redirect
        # Old test was:
        # data = rv.data.decode("utf-8")
        # ok_(ACCESS_IS_DENIED in data)
        self.assertEqual(rv.status_code, 404)

        # Reset My password
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        rv = client.get(
            f"/users/action/resetmypassword/{admin_user.id}", follow_redirects=True
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Reset Password Form", data)
        rv = client.post(
            "/resetmypassword/form",
            data=dict(password="password", conf_password="password"),
            follow_redirects=True,
        )
        self.assertEqual(rv.status_code, 200)
        self.browser_logout(client)
        self.browser_login(client, USERNAME_ADMIN, "password")
        rv = client.post(
            "/resetmypassword/form",
            data=dict(password=PASSWORD_ADMIN, conf_password=PASSWORD_ADMIN),
            follow_redirects=True,
        )
        self.assertEqual(rv.status_code, 200)

        # Reset Password Admin
        rv = client.get(
            f"/users/action/resetpasswords/{admin_user.id}", follow_redirects=True
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Reset Password Form", data)
        rv = client.post(
            "/resetmypassword/form",
            data=dict(password=PASSWORD_ADMIN, conf_password=PASSWORD_ADMIN),
            follow_redirects=True,
        )
        self.assertEqual(rv.status_code, 200)

    def test_sec_reset_password_default_complexity(self):
        """
        Test Security reset password with default complexity
        """
        client = self.app.test_client()
        self.app.config["FAB_PASSWORD_COMPLEXITY_ENABLED"] = True

        # Reset My password
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        rv = client.get("/users/action/resetmypassword/1", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Reset Password Form", data)
        rv = client.post(
            "/resetmypassword/form",
            data=dict(password="password", conf_password="password"),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")

        self.assertIn(PASSWORD_COMPLEXITY_ERROR, data)

        rv = client.post(
            "/resetmypassword/form",
            data=dict(password="PAssword123!", conf_password="PAssword123!"),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")

        self.assertNotIn(PASSWORD_COMPLEXITY_ERROR, data)

        # Revert changes
        self.app.config["FAB_PASSWORD_COMPLEXITY_ENABLED"] = False
        _ = client.post(
            "/resetmypassword/form",
            data=dict(password="password", conf_password="password"),
            follow_redirects=True,
        )

        self.browser_logout(client)

    def test_sec_reset_password_custom_complexity(self):
        """
        Test Security reset password with custom complexity
        """
        client = self.app.test_client()
        self.app.config["FAB_PASSWORD_COMPLEXITY_ENABLED"] = True
        self.app.config["FAB_PASSWORD_COMPLEXITY_VALIDATOR"] = custom_password_validator

        # Reset My password
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        rv = client.get("/users/action/resetmypassword/1", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Reset Password Form", data)
        rv = client.post(
            "/resetmypassword/form",
            data=dict(password="123", conf_password="123"),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")

        self.assertIn("Password must be password", data)

        rv = client.post(
            "/resetmypassword/form",
            data=dict(password="password", conf_password="password"),
            follow_redirects=True,
        )
        self.browser_logout(client)

    def test_register_user_with_role(self):
        """
        Test register user with role
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)

        # use all required params
        rv = client.get("/users/add", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Add User", data)
        rv = client.post(
            "/users/add",
            data=dict(
                first_name="first",
                last_name="last",
                username="from test 1-1",
                email="test1@fromtest1.com",
                roles=[1],
                groups=[],
                password="password",
                conf_password="password",
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Added Row", data)
        user = (
            self.appbuilder.session.query(User)
            .filter(User.username == "from test 1-1")
            .one_or_none()
        )
        self.appbuilder.session.delete(user)
        self.appbuilder.session.commit()

    def test_register_user_with_group(self):
        """
        Test register user with group
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        group = self.create_group(self.appbuilder)

        # use all required params
        rv = client.get("/users/add", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Add User", data)
        rv = client.post(
            "/users/add",
            data=dict(
                first_name="first",
                last_name="last",
                username="from test 1-1",
                email="test1@fromtest1.com",
                roles=[],
                groups=[group.id],
                password="password",
                conf_password="password",
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Added Row", data)
        user = (
            self.appbuilder.session.query(User)
            .filter(User.username == "from test 1-1")
            .one_or_none()
        )
        self.appbuilder.session.delete(user)
        self.appbuilder.session.delete(group)
        self.appbuilder.session.commit()

    def test_register_user_missing_roles_or_groups(self):
        """
        Test register user with missing roles or groups
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)

        # don't set roles or groups
        rv = client.get("/users/add", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Add User", data)
        rv = client.post(
            "/users/add",
            data=dict(
                first_name="first",
                last_name="last",
                username="from test 2-1",
                email="test2@fromtest2.com",
                roles=[],
                groups=[],
                password="password",
                conf_password="password",
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertNotIn("Added Row", data)
        self.assertIn("Either select a role or a group", data)
        self.browser_logout(client)

    def test_edit_user_with_role(self):
        """
        Test edit user with role
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)

        _tmp_user = self.create_user(
            self.appbuilder,
            "tmp_user",
            "password1",
            "",
            first_name="tmp",
            last_name="user",
            email="tmp@fab.org",
            role_names=["Admin"],
        )

        # use all required params
        rv = client.get(f"/users/edit/{_tmp_user.id}", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Edit User", data)
        rv = client.post(
            f"/users/edit/{_tmp_user.id}",
            data=dict(
                first_name=_tmp_user.first_name,
                last_name=_tmp_user.last_name,
                username=_tmp_user.username,
                email="changed@changed.org",
                roles=_tmp_user.roles[0].id,
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Changed Row", data)

        user = (
            self.appbuilder.session.query(User)
            .filter(User.username == _tmp_user.username)
            .one_or_none()
        )

        assert user.email == "changed@changed.org"
        self.appbuilder.session.delete(user)
        self.appbuilder.session.commit()

    def test_edit_user_with_group(self):
        """
        Test edit user with group
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)

        _tmp_user = self.create_user(
            self.appbuilder,
            "tmp_user",
            "password1",
            "",
            first_name="tmp",
            last_name="user",
            email="tmp@fab.org",
            role_names=["Admin"],
        )

        # use all required params
        rv = client.get(f"/users/edit/{_tmp_user.id}", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Edit User", data)
        rv = client.post(
            f"/users/edit/{_tmp_user.id}",
            data=dict(
                first_name=_tmp_user.first_name,
                last_name=_tmp_user.last_name,
                username=_tmp_user.username,
                email="changed@changed.org",
                roles=_tmp_user.roles[0].id,
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Changed Row", data)

        user = (
            self.appbuilder.session.query(User)
            .filter(User.username == _tmp_user.username)
            .one_or_none()
        )

        assert user.email == "changed@changed.org"
        self.appbuilder.session.delete(user)
        self.appbuilder.session.commit()

    def test_edit_user_email_validation(self):
        """
        Test edit user with email not null validation
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)

        read_ony_user: User = (
            self.appbuilder.session.query(User)
            .filter(User.username == USERNAME_READONLY)
            .one_or_none()
        )

        # use all required params
        rv = client.get(f"/users/edit/{read_ony_user.id}", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Edit User", data)
        rv = client.post(
            f"/users/edit/{read_ony_user.id}",
            data=dict(
                first_name=read_ony_user.first_name,
                last_name=read_ony_user.last_name,
                username=read_ony_user.username,
                email=None,
                roles=read_ony_user.roles[0].id,
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("This field is required", data)

    def test_edit_user_db_fail(self):
        """
        Test edit user with DB fail
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)

        read_ony_user: User = (
            self.appbuilder.session.query(User)
            .filter(User.username == USERNAME_READONLY)
            .one_or_none()
        )

        # use all required params
        rv = client.get(f"/users/edit/{read_ony_user.id}", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Edit User", data)

        with patch.object(self.appbuilder.session, "merge") as mock_merge:
            with patch.object(self.appbuilder.sm, "has_access", return_value=True) as _:
                mock_merge.side_effect = Exception("BANG!")

                rv = client.post(
                    f"/users/edit/{read_ony_user.id}",
                    data=dict(
                        first_name=read_ony_user.first_name,
                        last_name=read_ony_user.last_name,
                        username=read_ony_user.username,
                        email="changed@changed.org",
                        roles=read_ony_user.roles[0].id,
                    ),
                    follow_redirects=True,
                )

                data = rv.data.decode("utf-8")
                self.assertIn("Database Error", data)

    def test_add_group(self):
        """
        Test add group
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)

        # use all required params
        rv = client.get("/groups/add", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Add Group", data)
        rv = client.post(
            "/groups/add",
            data=dict(
                name="group1",
                label="group1",
                description="some description",
                roles=[1],
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Added Row", data)
        group = (
            self.appbuilder.session.query(Group)
            .filter(Group.name == "group1")
            .one_or_none()
        )
        self.appbuilder.session.delete(group)
        self.appbuilder.session.commit()

    def test_add_group_unique_name(self):
        """
        Test add group unique name
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        group = self.create_group(self.appbuilder)
        self.appbuilder.session.refresh(group)
        rv = client.get("/groups/add", follow_redirects=True)
        data = rv.data.decode("utf-8")
        self.assertIn("Add Group", data)
        rv = client.post(
            "/groups/add",
            data=dict(
                name=group.name,
                label="group1",
                description="some description",
                roles=[1],
            ),
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Already exists.", data)
        self.appbuilder.session.delete(group)
        self.appbuilder.session.commit()

    def test_delete_group(self):
        """
        Test delete group
        """
        client = self.app.test_client()
        _ = self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        group = self.create_group(self.appbuilder)
        rv = client.post(
            f"/groups/delete/{group.id}",
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("Deleted Row", data)
        assert self.appbuilder.sm.find_group(name="group1") is None

    def test_delete_group_with_users(self):
        """
        Test delete group with users
        """
        client = self.app.test_client()
        self.browser_login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        group = self.create_group(self.appbuilder)
        user = self.create_user(
            self.appbuilder,
            "test_user",
            "password",
            None,
            first_name="test",
            last_name="user",
            email="test_user@fab.org",
            group_names=["group1"],
        )
        rv = client.post(
            f"/groups/delete/{group.id}",
            follow_redirects=True,
        )
        data = rv.data.decode("utf-8")
        self.assertIn("User(s) exists in the group, cannot delete", data)
        assert self.appbuilder.sm.find_group(name="group1") is not None
        self.appbuilder.session.delete(user)
        self.appbuilder.session.delete(group)
        self.appbuilder.session.commit()



import logging

from flask import Flask
from .views import (
    ContactModelView,
    GroupModelView,
    ContactChartView,
    ContactTimeChartView,
)
from .extensions import appbuilder, db


logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)


def create_app() -> Flask:
    app = Flask(__name__)
    app.config.from_object("config")
    with app.app_context():
        db.init_app(app)
        appbuilder.init_app(app, db.session)
        appbuilder.add_view(
            GroupModelView,
            "List Groups",
            icon="fa-folder-open-o",
            category="Contacts",
            category_icon="fa-envelope",
        )
        appbuilder.add_view(
            ContactModelView, "List Contacts", icon="fa-envelope", category="Contacts"
        )
        appbuilder.add_separator("Contacts")
        appbuilder.add_view(
            ContactChartView, "Contacts Chart", icon="fa-dashboard", category="Contacts"
        )
        appbuilder.add_view(
            ContactTimeChartView,
            "Contacts Birth Chart",
            icon="fa-dashboard",
            category="Contacts",
        )
    return app


# For backward compatibility
app = create_app()



import logging

from flask import Flask
from .views import (
    DepartmentView,
    EmployeeView,
    FunctionView,
    EmployeeHistoryView,
    BenefitView,
)
from .extensions import appbuilder, db


logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)


def create_app() -> Flask:
    app = Flask(__name__)
    app.config.from_object("config")
    with app.app_context():
        db.init_app(app)
        appbuilder.init_app(app, db.session)

        appbuilder.add_view_no_menu(EmployeeHistoryView, "EmployeeHistoryView")
        appbuilder.add_view(
            EmployeeView, "Employees", icon="fa-folder-open-o", category="Company"
        )
        appbuilder.add_separator("Company")
        appbuilder.add_view(
            DepartmentView, "Departments", icon="fa-folder-open-o", category="Company"
        )
        appbuilder.add_view(
            FunctionView, "Functions", icon="fa-folder-open-o", category="Company"
        )
        appbuilder.add_view(
            BenefitView, "Benefits", icon="fa-folder-open-o", category="Company"
        )

    return app


# For backward compatibility
app = create_app()



import os

from flask import Flask
from flask_appbuilder import AppBuilder
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder.urltools import get_filter_args
from flask_appbuilder.utils.legacy import get_sqla_class
from tests.base import FABTestCase
from tests.sqla.models import Model1


class FlaskTestCase(FABTestCase):
    def setUp(self):
        self.app = Flask(__name__)
        self.basedir = os.path.abspath(os.path.dirname(__file__))
        self.app.config.from_object("tests.config_api")
        self.ctx = self.app.app_context()
        self.ctx.push()
        SQLA = get_sqla_class()
        self.db = SQLA(self.app)
        self.appbuilder = AppBuilder(self.app, self.db.session)

    def tearDown(self):
        self.ctx.pop()

    def test_get_filter_args_allow_one(self):
        datamodel = SQLAInterface(Model1)
        with self.app.test_request_context("/users/list?_flt_1_field_string=a"):
            filters = datamodel.get_filters(["field_string", "field_integer"])
            get_filter_args(filters)
            assert filters.values == [["a"]]

    def test_get_filter_args_allow_multiple(self):
        datamodel = SQLAInterface(Model1)
        with self.app.test_request_context(
            "/users/list?_flt_1_field_string=a&_flt_1_field_integer=2"
        ):
            filters = datamodel.get_filters(["field_string", "field_integer"])
            get_filter_args(filters)
            assert filters.values in ([["a"], ["2"]], [["2"], ["a"]])

    def test_get_filter_args_disallow(self):
        datamodel = SQLAInterface(Model1)
        with self.app.test_request_context("/users/list?_flt_1_field_float=1.0"):
            filters = datamodel.get_filters(["field_string", "field_integer"])
            get_filter_args(filters)
            assert filters.values == []

    def test_get_filter_args_disallow_off(self):
        datamodel = SQLAInterface(Model1)
        with self.app.test_request_context("/users/list?_flt_1_field_float=1.0"):
            filters = datamodel.get_filters(["field_string", "field_integer"])
            get_filter_args(filters, disallow_if_not_in_search=False)
            assert filters.values == [["1.0"]]

    def test_get_filter_args_invalid_index(self):
        datamodel = SQLAInterface(Model1)
        with self.app.test_request_context("/users/list?_flt_a_field_string=a"):
            filters = datamodel.get_filters(["field_string", "field_integer"])
            get_filter_args(filters)
            assert filters.values == []



import os

from flask import Flask
from flask_appbuilder import AppBuilder
from flask_appbuilder.utils.legacy import get_sqla_class

# Get the SQLA class
SQLA = get_sqla_class()

# Create app factory
def create_app():
    app = Flask(__name__)
    
    basedir = os.path.abspath(os.path.dirname(__file__))
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///" + os.path.join(basedir, "app.db")
    app.config["CSRF_ENABLED"] = True
    app.config["SECRET_KEY"] = "thisismyscretkey"
    
    db = SQLA()
    appbuilder = AppBuilder()
    
    with app.app_context():
        db.init_app(app)
        appbuilder.init_app(app, db.session)
    
    return app

# Create the app
app = create_app()

if __name__ == "__main__":
    app.run(host="0.0.0.0", port=8080, debug=True)



from flask import Markup, url_for
from flask_appbuilder import Model
from flask_appbuilder.filemanager import ImageManager
from flask_appbuilder.models.mixins import ImageColumn
from flask_appbuilder.security.sqla.models import User
from sqlalchemy import Column, Float, ForeignKey, Integer, String, Text
from sqlalchemy.orm import relationship


class ProductType(Model):
    __tablename__ = "product_type"
    id = Column(Integer, primary_key=True)
    name = Column(String(50), unique=True, nullable=False)

    def __repr__(self):
        return self.name


class Product(Model):
    __tablename__ = "product"
    id = Column(Integer, primary_key=True)
    name = Column(String(50), unique=True, nullable=False)
    price = Column(Float, nullable=False)
    photo = Column(ImageColumn)
    description = Column(Text())
    product_type_id = Column(Integer, ForeignKey("product_type.id"), nullable=False)
    product_type = relationship("ProductType")

    def photo_img(self):
        im = ImageManager()
        productPubViewUrl = url_for("ProductPubView.show", pk=str(self.id))
        if self.photo:
            photoUrl = im.get_url(self.photo)
            return Markup(
                f'<a href="{productPubViewUrl}" class="thumbnail"><img src="{photoUrl}" '
                f'alt="Photo" class="img-rounded img-responsive"></a>'
            )
        else:
            return Markup(
                f'<a href="{productPubViewUrl}"" class="thumbnail"><img src="//:0" '
                f'alt="Photo" class="img-responsive">' '</a>'
            )

    def price_label(self):
        return Markup("Price:<strong> {} </strong>".format(self.price))

    def __repr__(self):
        return self.name


class Client(User):
    __tablename__ = "ab_user"
    extra = Column(String(50), unique=True, nullable=False)



import logging
import os

from flask_appbuilder import AppBuilder, IndexView
from flask_appbuilder.utils.legacy import get_sqla_class
from tests.base import FABTestCase

log = logging.getLogger(__name__)


class CustomIndexView(IndexView):
    index_template = "templates/custom_index.html"


class FlaskTestCase(FABTestCase):
    def setUp(self):
        from flask import Flask

        self.app = Flask(__name__, template_folder=".")
        self.basedir = os.path.abspath(os.path.dirname(__file__))
        self.app.config.from_object("tests.config_api")
        self.app.config[
            "FAB_INDEX_VIEW"
        ] = "tests.test_custom_indexview.CustomIndexView"

    def test_custom_indexview(self):
        """
        Test custom index view.
        """
        uri = "/"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            client = self.app.test_client()
            rv = client.get(uri)

            self.assertEqual(rv.status_code, 200)
            data = rv.data.decode("utf-8")
            self.assertIn("This is a custom index view.", data)



import logging
import os
from typing import List
import unittest
from unittest.mock import Mock

from flask import Flask
from flask_appbuilder import AppBuilder
from flask_appbuilder.security.manager import AUTH_LDAP
from flask_appbuilder.security.sqla.models import User
from flask_appbuilder.utils.legacy import get_sqla_class
import jinja2
import ldap
from tests.const import USERNAME_ADMIN, USERNAME_READONLY
from tests.fixtures.users import create_default_users

logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)
log = logging.getLogger(__name__)


class LDAPSearchTestCase(unittest.TestCase):
    def setUp(self):
        # start MockLdap
        # start Flask
        self.app = Flask(__name__)
        self.app.jinja_env.undefined = jinja2.StrictUndefined
        self.app.config["SQLALCHEMY_DATABASE_URI"] = os.environ.get(
            "SQLALCHEMY_DATABASE_URI", "sqlite:///"
        )
        self.app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
        self.app.config["AUTH_TYPE"] = AUTH_LDAP
        self.app.config["AUTH_LDAP_SERVER"] = "ldap://localhost:1389/"
        self.app.config["AUTH_LDAP_UID_FIELD"] = "uid"
        self.app.config["AUTH_LDAP_FIRSTNAME_FIELD"] = "givenName"
        self.app.config["AUTH_LDAP_LASTNAME_FIELD"] = "sn"
        self.app.config["AUTH_LDAP_EMAIL_FIELD"] = "mail"

    def tearDown(self):
        # Remove test user
        with self.app.app_context():
            # Remove test user
            user_alice = self.appbuilder.sm.find_user("alice")
            if user_alice:
                self.appbuilder.session.delete(user_alice)
                self.appbuilder.session.commit()
            user_natalie = self.appbuilder.sm.find_user("natalie")
            if user_natalie:
                self.appbuilder.session.delete(user_natalie)
                self.appbuilder.session.commit()

    def assertOnlyDefaultUsers(self):
        users = self.appbuilder.sm.get_all_users()
        user_names = sorted([user.username for user in users])
        self.assertEqual(user_names, [USERNAME_READONLY, USERNAME_ADMIN])

    def assertUserContainsRoles(self, user: User, role_names: List[str]):
        user_role_names = sorted([role.name for role in user.roles])
        self.assertListEqual(user_role_names, sorted(role_names))

    # ----------------
    # Unit Tests
    # ----------------
    def test___search_ldap(self):
        """
        LDAP: test `_search_ldap` method
        """
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # prepare `con` object
            con = ldap.initialize("ldap://localhost:1389/")
            sm._ldap_bind_indirect(ldap, con)

            # run `_search_ldap` method
            user_dn, user_attributes = sm._search_ldap(ldap, con, "alice")

            # validate - search returned expected data
            self.assertEqual(user_dn, "cn=alice,ou=users,dc=example,dc=org")
            self.assertEqual(user_attributes["givenName"], [b"Alice"])
            self.assertEqual(user_attributes["sn"], [b"Doe"])
            self.assertEqual(user_attributes["mail"], [b"alice@example.org"])

    def test___search_ldap_filter(self):
        """
        LDAP: test `_search_ldap` method (with AUTH_LDAP_SEARCH_FILTER)
        """
        # MockLdap needs non-bytes for search filters, so we patch `memberOf`
        # to a string, only for this test

        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config[
            "AUTH_LDAP_SEARCH_FILTER"
        ] = "(memberOf=cn=staff,ou=groups,dc=example,dc=org)"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # prepare `con` object
            con = ldap.initialize("ldap://localhost:1389/")
            sm._ldap_bind_indirect(ldap, con)

            # run `_search_ldap` method
            user_dn, user_attributes = sm._search_ldap(ldap, con, "alice")

            # validate - search returned expected data
            self.assertEqual(user_dn, "cn=alice,ou=users,dc=example,dc=org")
            self.assertEqual(user_attributes["givenName"], [b"Alice"])
            self.assertEqual(user_attributes["sn"], [b"Doe"])
            self.assertEqual(user_attributes["mail"], [b"alice@example.org"])

    def test___search_ldap_with_search_referrals(self):
        """
        LDAP: test `_search_ldap` method w/returned search referrals
        """
        self.app.config["AUTH_LDAP_BIND_USER"] = "uid=admin,ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            user_alice = (
                "cn=alice,ou=users,dc=example,dc=org",
                {
                    "uid": ["alice"],
                    "userPassword": ["alice_password"],
                    "memberOf": [b"cn=staff,ou=groups,o=test"],
                    "givenName": [b"Alice"],
                    "sn": [b"Doe"],
                    "mail": [b"alice@example.org"],
                },
            )
            # run `_search_ldap` method w/mocked ldap connection
            mock_con = Mock()
            mock_con.search_s.return_value = [
                (
                    None,
                    [
                        "ldap://ForestDnsZones.mycompany.com/"
                        "DC=ForestDnsZones,DC=mycompany,DC=com"
                    ],
                ),
                user_alice,
                (None, ["ldap://mycompany.com/CN=Configuration,DC=mycompany,DC=com"]),
            ]
            user_dn, user_attributes = sm._search_ldap(ldap, mock_con, "alice")

            # validate - search returned expected data
            self.assertEqual(user_dn, user_alice[0])
            self.assertEqual(user_attributes["givenName"], user_alice[1]["givenName"])
            self.assertEqual(user_attributes["sn"], user_alice[1]["sn"])
            self.assertEqual(user_attributes["mail"], user_alice[1]["mail"])

            mock_con.search_s.assert_called()

    def test__missing_credentials(self):
        """
        LDAP: test login flow for - missing credentials
        """
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # validate - login failure (missing username)
            self.assertIsNone(sm.auth_user_ldap(None, "password"))
            self.assertIsNone(sm.auth_user_ldap("", "password"))

            # validate - login failure (missing password)
            self.assertIsNone(sm.auth_user_ldap("username", None))
            self.assertIsNone(sm.auth_user_ldap("username", ""))

            # validate - login failure (missing username/password)
            self.assertIsNone(sm.auth_user_ldap(None, None))
            self.assertIsNone(sm.auth_user_ldap("", None))
            self.assertIsNone(sm.auth_user_ldap("", ""))
            self.assertIsNone(sm.auth_user_ldap(None, ""))

            # validate - no users were created
            self.assertOnlyDefaultUsers()

    def test__active_user(self):
        """
        LDAP: test login flow for - active user
        """
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.com",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # set user inactive
            new_user.active = True

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was not allowed to log in
            self.assertIsNotNone(user)

    def test__inactive_user(self):
        """
        LDAP: test login flow for - inactive user
        """
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.com",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # set user inactive
            new_user.active = False

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was not allowed to log in
            self.assertIsNone(user)

    def test__multi_group_user_mapping_to_same_role(self):
        """
        LDAP: test login flow for - user in multiple groups mapping to same role
        """
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin"],
            "cn=readers,ou=groups,dc=example,dc=org": ["User"],
        }
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("natalie", "natalie_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertUserContainsRoles(user, ["Public", "User"])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Natalie")
            self.assertEqual(user.last_name, "Smith")
            self.assertEqual(user.email, "natalie@example.org")

    def test__direct_bind__unregistered(self):
        """
        LDAP: test login flow for - direct bind - unregistered user
        """
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the AUTH_USER_REGISTRATION_ROLE role
            self.assertEqual(user.roles, [sm.find_role("Public")])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.org")

    def test__direct_bind__unregistered__no_self_register(self):
        """
        LDAP: test login flow for - direct bind - unregistered user - no self-registration
        """
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_USER_REGISTRATION"] = False
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was not allowed to log in
            self.assertIsNone(user)

            # validate - no users were registered
            self.assertOnlyDefaultUsers()

    def test__direct_bind__unregistered__no_search(self):
        """
        LDAP: test login flow for - direct bind - unregistered user - no ldap search
        """
        self.app.config["AUTH_LDAP_SEARCH"] = None
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was NOT allowed to log in
            # (because registration requires search)
            self.assertIsNone(user)

    def test__direct_bind__registered(self):
        """
        LDAP: test login flow for - direct bind - registered user
        """
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

    def test__direct_bind__registered__no_search(self):
        """
        LDAP: test login flow for - direct bind - registered user - no ldap search
        """
        self.app.config["AUTH_LDAP_SEARCH"] = None
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in (because they are already registered)
            self.assertIsInstance(user, sm.user_model)

    def test__indirect_bind__unregistered(self):
        """
        LDAP: test login flow for - indirect bind - unregistered user
        """
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the AUTH_USER_REGISTRATION_ROLE role
            self.assertListEqual(user.roles, [sm.find_role("Public")])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.org")

    def test__indirect_bind__unregistered__no_self_register(self):
        """
        LDAP: test login flow for - indirect bind - unregistered user - no self-registration
        """  # noqa
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_USER_REGISTRATION"] = False
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was not allowed to log in
            self.assertIsNone(user)

            # validate - no users were registered
            self.assertOnlyDefaultUsers()

    def test__indirect_bind__unregistered__no_search(self):
        """
        LDAP: test login flow for - indirect bind - unregistered user - no ldap search
        """
        self.app.config["AUTH_LDAP_SEARCH"] = None
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was NOT allowed to log in
            # (because indirect bind requires search)
            self.assertIsNone(user)

    def test__indirect_bind__registered(self):
        """
        LDAP: test login flow for - indirect bind - registered user
        """
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

    def test__indirect_bind__registered__no_search(self):
        """
        LDAP: test login flow for - indirect bind - registered user - no ldap search
        """
        self.app.config["AUTH_LDAP_SEARCH"] = None
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was NOT allowed to log in
            # (because indirect bind requires search)
            self.assertIsNone(user)

    def test__direct_bind__unregistered__single_role(self):
        """
        LDAP: test login flow for - direct bind - unregistered user - single role mapping
        """
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin"]
        }
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertUserContainsRoles(user, ["Admin", "Public"])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.org")

    def test__direct_bind__unregistered__multi_role(self):
        """
        LDAP: test login flow for - direct bind - unregistered user - multi role mapping
        """
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin", "User"]
        }
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertUserContainsRoles(user, ["Admin", "Public", "User"])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.org")

    def test__direct_bind__registered__multi_role__no_role_sync(self):
        """
        LDAP: test login flow for - direct bind - registered user - multi role mapping - no login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin", "User"]
        }
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = False
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given no roles
            self.assertListEqual(user.roles, [])

    def test__direct_bind__registered__multi_role__with_role_sync(self):
        """
        LDAP: test login flow for - direct bind - registered user - multi role mapping - with login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin", "User"]
        }
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = True
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given the correct roles
            self.assertUserContainsRoles(user, ["Admin", "User"])

    def test__indirect_bind__unregistered__single_role(self):
        """
        LDAP: test login flow for - indirect bind - unregistered user - single role mapping
        """  # noqa
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["User"]
        }
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertUserContainsRoles(user, ["Public", "User"])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.org")

    def test__indirect_bind__unregistered__multi_role(self):
        """
        LDAP: test login flow for - indirect bind - unregistered user - multi role mapping
        """
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin", "User"]
        }
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # validate - user was given the correct roles
            self.assertUserContainsRoles(user, ["User", "Public", "Admin"])

            # validate - user was given the correct attributes (read from LDAP)
            self.assertEqual(user.first_name, "Alice")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "alice@example.org")

    def test__indirect_bind__registered__multi_role__no_role_sync(self):
        """
        LDAP: test login flow for - indirect bind - registered user - multi role mapping - no login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin", "User"]
        }
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = False
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given no roles
            self.assertListEqual(user.roles, [])

    def test__indirect_bind__registered__multi_role__with_role_sync(self):
        """
        LDAP: test login flow for - indirect bind - registered user - multi role mapping - with login role-sync
        """  # noqa
        self.app.config["AUTH_ROLES_MAPPING"] = {
            "cn=staff,ou=groups,dc=example,dc=org": ["Admin", "User"]
        }
        self.app.config["AUTH_ROLES_SYNC_AT_LOGIN"] = True
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_USER"] = "cn=admin,dc=example,dc=org"
        self.app.config["AUTH_LDAP_BIND_PASSWORD"] = "admin_password"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            sm = self.appbuilder.sm
            create_default_users(self.appbuilder.session)

            # add User role
            sm.add_role("User")

            # validate - no users are registered
            self.assertOnlyDefaultUsers()

            # register a user
            new_user = sm.add_user(  # noqa
                username="alice",
                first_name="Alice",
                last_name="Doe",
                email="alice@example.org",
                role=[],
            )

            # validate - user was registered
            self.assertEqual(len(sm.get_all_users()), 3)

            # attempt login
            user = sm.auth_user_ldap("alice", "alice_password")

            # validate - user was allowed to log in
            self.assertIsInstance(user, sm.user_model)

            # validate - user was given the correct roles
            self.assertUserContainsRoles(user, ["User", "Admin"])

    def test_login_failed_keep_next_url(self):
        """
        LDAP: Keeping next url after failed login attempt
        """
        self.app.config["AUTH_LDAP_SEARCH"] = "ou=users,dc=example,dc=org"
        self.app.config[
            "AUTH_LDAP_USERNAME_FORMAT"
        ] = "cn=%s,ou=users,dc=example,dc=org"
        self.app.config["AUTH_USER_REGISTRATION"] = True
        self.app.config["AUTH_USER_REGISTRATION_ROLE"] = "Public"
        self.app.config["WTF_CSRF_ENABLED"] = False
        self.app.config["SECRET_KEY"] = "thisismyscretkey"
        with self.app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(self.app)
            self.appbuilder = AppBuilder(self.app, db.session)
            client = self.app.test_client()
            client.get("/logout/")

            response = client.post(
                "/login/?next=/users/userinfo/",
                data=dict(username="natalie", password="wrong_natalie_password"),
                follow_redirects=False,
            )
            response = client.post(
                response.location,
                data=dict(username="natalie", password="natalie_password"),
                follow_redirects=False,
            )
            assert response.location == "/users/userinfo/"



from datetime import datetime
import logging
import random

from flask_appbuilder.extensions import db

from app import create_app
from app.extensions import appbuilder
from app.models import ContactGroup, Gender, Contact, Company


log = logging.getLogger(__name__)



def get_random_name(names_list, size=1):
    name_lst = [
        names_list[random.randrange(0, len(names_list))].decode("utf-8").capitalize()
        for i in range(0, size)
    ]
    return " ".join(name_lst)


app = create_app()

def upsert_test_data():
    company1 = Company(name="Company 1")
    company2 = Company(name="Company 2")
    try:
        db.session.add(company1)
        db.session.add(company2)
        db.session.commit()
    except Exception as e:
        log.error("Group creation error: %s", e)
        db.session.rollback()
        exit(1)


    role_admin = appbuilder.sm.find_role(appbuilder.sm.auth_role_admin)

    user1 = appbuilder.sm.add_user(
        "user1_company1", "user1", "test", "user1@company1.com", role_admin, "password"
    )
    user2 = appbuilder.sm.add_user(
        "user1_company2", "user1", "test", "user1@company2.com", role_admin, "password"
    )
    user3 = appbuilder.sm.add_user(
        "user2_company2", "user2", "test", "user2@company2.com", role_admin, "password"
    )
    user1.company = company1
    user2.company = company2
    user3.company = company2
    db.session.merge(user1)
    db.session.merge(user2)
    db.session.merge(user3)
    db.session.commit()

    try:
        db.session.add(ContactGroup(name="Friends"))
        db.session.add(ContactGroup(name="Family"))
        db.session.add(ContactGroup(name="Work"))
        db.session.commit()
    except Exception as e:
        log.error("Group creation error: %s", e)
        db.session.rollback()
        exit(1)

    try:
        db.session.add(Gender(name="Male"))
        db.session.add(Gender(name="Female"))
        db.session.commit()
    except Exception as e:
        log.error("Gender creation error: %s", e)
        db.session.rollback()
        exit(1)

    f = open("NAMES.DIC", "rb")
    names_list = [x.strip() for x in f.readlines()]

    f.close()

    j = 1
    for i in range(1, 100):
        c = Contact()
        c.name = get_random_name(names_list, random.randrange(2, 6))
        c.address = "Street " + names_list[random.randrange(0, len(names_list))].decode(
            "utf-8"
        )
        c.personal_phone = random.randrange(1111111, 9999999)
        c.personal_celphone = random.randrange(1111111, 9999999)
        c.contact_group_id = random.randrange(1, 4)
        c.gender_id = random.randrange(1, 3)
        year = random.choice(range(1900, 2012))
        month = random.choice(range(1, 12))
        day = random.choice(range(1, 28))
        c.birthday = datetime(year, month, day)
        c.changed_on = datetime.now()
        c.created_on = datetime.now()
        if j == 1:
            j += 1
            _user = user1
        elif j == 2:
            j += 1
            _user = user2
        else:
            j = 1
            _user = user3
        c.created_by = _user
        c.changed_by = _user

        db.session.add(c)
        try:
            db.session.commit()
            print("inserted", c)
        except Exception as e:
            log.error("Contact creation error: %s", e)
            db.session.rollback()

with app.app_context():
    upsert_test_data()



from flask import g
from flask_appbuilder import ModelView
from flask_appbuilder.models.filters import BaseFilter
from flask_appbuilder.models.sqla.filters import get_field_setup_query
from flask_appbuilder.models.sqla.interface import SQLAInterface

from . import appbuilder, db
from .models import Company, Contact, ContactGroup
from .sec_views import UserDBModelView


class FilterInManyFunction(BaseFilter):
    name = "Filter view where field is in a list returned by a function"

    def apply(self, query, func):
        query, field = get_field_setup_query(query, self.model, self.column_name)
        return query.filter(field.any(Company.id.in_(func())))


def get_user_companies():
    return set([company.id for company in g.user.companies])
    return g.user.companies


class ContactModelView(ModelView):
    datamodel = SQLAInterface(Contact)
    list_columns = [
        "name",
        "personal_celphone",
        "birthday",
        "contact_group.name",
        "created_by",
    ]
    add_columns = [
        "name",
        "address",
        "birthday",
        "personal_phone",
        "personal_celphone",
        "contact_group",
        "gender",
    ]
    edit_columns = [
        "name",
        "address",
        "birthday",
        "personal_phone",
        "personal_celphone",
        "contact_group",
        "gender",
    ]
    base_order = ("name", "asc")
    base_filters = [["created_by.companies", FilterInManyFunction, get_user_companies]]


class GroupModelView(ModelView):
    datamodel = SQLAInterface(ContactGroup)
    related_views = [ContactModelView]


class CompanyModelView(ModelView):
    datamodel = SQLAInterface(Company)
    list_columns = ["name", "myuser"]
    related_views = [UserDBModelView]


db.create_all()
appbuilder.add_view(CompanyModelView, "Companys", icon="fa-folder-open-o")
appbuilder.add_view(
    GroupModelView,
    "List Groups",
    icon="fa-folder-open-o",
    category="Contacts",
    category_icon="fa-envelope",
)
appbuilder.add_view(
    ContactModelView, "List Contacts", icon="fa-envelope", category="Contacts"
)
appbuilder.add_separator("Contacts")



import logging
import os

from tests.base import FABTestCase
from tests.fixtures.addon_manager import DummyAddOnManager


log = logging.getLogger(__name__)


class FlaskTestCase(FABTestCase):
    def setUp(self):
        from flask import Flask
        from flask_appbuilder import AppBuilder
        from flask_appbuilder.utils.legacy import get_sqla_class

        self.app = Flask(__name__)
        self.basedir = os.path.abspath(os.path.dirname(__file__))
        self.app.config.from_object("tests.config_api")
        self.app.config["ADDON_MANAGERS"] = [
            "tests.fixtures.addon_manager.DummyAddOnManager"
        ]
        self.ctx = self.app.app_context()
        self.ctx.push()
        SQLA = get_sqla_class()
        self.db = SQLA(self.app)
        self.appbuilder = AppBuilder(self.app, self.db.session)

    def tearDown(self):
        self.ctx.pop()
        self.appbuilder = None
        self.app = None
        self.db = None

    def test_addon_import(self):
        self.assertIsInstance(
            self.appbuilder.addon_managers[
                "tests.fixtures.addon_manager.DummyAddOnManager"
            ],
            DummyAddOnManager,
        )

    def test_addon_register_views(self):
        client = self.app.test_client()
        rv = client.get("/dummy/method1/test1")
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv.data.decode("utf-8"), "test1")



import os

basedir = os.path.abspath(os.path.dirname(__file__))

SQLALCHEMY_DATABASE_URI = os.environ.get(
    "SQLALCHEMY_DATABASE_URI"
) or "sqlite:///" + os.path.join(basedir, "app.db")


FAB_ADD_SECURITY_API = True
SECRET_KEY = "thisismyscretkey"
SQLALCHEMY_TRACK_MODIFICATIONS = False
WTF_CSRF_ENABLED = False
FAB_API_SWAGGER_UI = True
FAB_ROLES = {
    "ReadOnly": [
        [".*", "can_get"],
        [".*", "can_info"],
        [".*", "can_list"],
        [".*", "can_show"],
    ]
}



import glob
import json
import logging
import os
import tempfile
from typing import List
from unittest.mock import ANY, patch

from click.testing import CliRunner
from flask import Flask
from flask_appbuilder import AppBuilder
from flask_appbuilder.cli import (
    cast_int_like_to_int,
    create_app,
    # create_permissions,
    # create_user,
    export_roles,
    import_roles,
    # list_users,
    # list_views,
    # reset_password,
)
from flask_appbuilder.utils.legacy import get_sqla_class
from tests.base import FABTestCase

logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)
log = logging.getLogger(__name__)

APP_DIR = "myapp"


class FlaskTestCase(FABTestCase):
    def setUp(self):
        pass

    def tearDown(self):
        log.debug("TEAR DOWN")

    def test_create_app_invalid_secret_key(self):
        os.environ["FLASK_APP"] = "app:app"
        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(
                create_app,
                [
                    f"--name={APP_DIR}",
                    "--secret-key=SHORT_SECRET",
                ],
            )
            self.assertIn("Invalid value for '--secret-key'", result.output)

    test_create_app_invalid_secret_key.needs_inet = True

    def test_create_app(self):
        """
        Test create app, create-user
        """
        os.environ["FLASK_APP"] = "app:app"
        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(
                create_app,
                [
                    f"--name={APP_DIR}",
                    f"--secret-key={10*'SECRET'}",
                ],
            )
            self.assertIn("Downloaded the skeleton app, good coding!", result.output)
            # os.chdir(APP_DIR)
            # result = runner.invoke(
            #     create_user,
            #     [
            #         "--username=bob",
            #         "--role=Public",
            #         "--firstname=Bob",
            #         "--lastname=Smith",
            #         "--email=bob@fab.com",
            #         "--password=foo",
            #     ],
            # )
            # log.info(result.output)
            # self.assertIn("User bob created.", result.output)
            #
            # result = runner.invoke(list_users, [])
            # self.assertIn("bob", result.output)
            #
            # runner.invoke(create_permissions, [])
            #
            # runner.invoke(reset_password, ["--username=bob", "--password=bar"])

    test_create_app.needs_inet = True

    # def test_list_views(self):
    #     """
    #     CLI: Test list views
    #     """
    #     os.environ["FLASK_APP"] = "app:app"
    #     runner = CliRunner()
    #     with runner.isolated_filesystem():
    #         result = runner.invoke(
    #             create_app,
    #             [
    #                 f"--name={APP_DIR}",
    #                 "--engine=SQLAlchemy",
    #                 f"--secret-key={10*'SECRET'}",
    #             ],
    #         )
    #         self.assertIn("Downloaded the skeleton app, good coding!", result.output)
    #         os.chdir(APP_DIR)
    #
    #         result = runner.invoke(list_views, [])
    #         self.assertIn("List of registered views", result.output)
    #         self.assertIn(" Route:/api/v1/security", result.output)
    #
    # test_list_views.needs_inet = True

    def test_cast_int_like_to_int(self):
        scenarii = {
            -1: -1,
            0: 0,
            1: 1,
            "-1": -1,
            "0": 0,
            "1": 1,
            "+1": 1,
            "foo": "foo",
            None: None,
        }

        for input, expected_output in scenarii.items():
            self.assertEqual(cast_int_like_to_int(input), expected_output)


class SQLAlchemyImportExportTestCase(FABTestCase):
    def setUp(self):
        basedir = os.path.abspath(os.path.dirname(__file__))

        with open(os.path.join(basedir, "data/roles.json"), "r") as fd:
            self.expected_roles = json.loads(fd.read())

    def test_export_roles(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            app = Flask("src_app")
            app.config.from_object("tests.config_security_cli")
            with app.app_context():
                SQLA = get_sqla_class()
                db = SQLA(app)
                app_builder = AppBuilder(app, db.session)  # noqa: F841
                cli_runner = app.test_cli_runner()

                path = os.path.join(tmp_dir, "roles.json")

                export_result = cli_runner.invoke(export_roles, [f"--path={path}"])

                self.assertEqual(export_result.exit_code, 0)
                self.assertTrue(os.path.exists(path))

                with open(path, "r") as fd:
                    resulting_roles: List = json.loads(fd.read())
                # sort to get consistent order and consistent tests
                self.expected_roles = sorted(
                    self.expected_roles, key=lambda role_name: role_name["name"]
                )
                resulting_roles = sorted(
                    resulting_roles, key=lambda role_name: role_name["name"]
                )
                for expected_role in self.expected_roles:
                    match = [
                        r for r in resulting_roles if r["name"] == expected_role["name"]
                    ]
                    self.assertTrue(match)
                    resulting_role = match[0]
                    resulting_role_permission_view_menus = {
                        (pvm["permission"]["name"], pvm["view_menu"]["name"])
                        for pvm in resulting_role["permissions"]
                    }
                    expected_role_permission_view_menus = {
                        (pvm["permission"]["name"], pvm["view_menu"]["name"])
                        for pvm in expected_role["permissions"]
                    }
                    self.assertEqual(
                        resulting_role_permission_view_menus,
                        expected_role_permission_view_menus,
                    )

    def test_export_roles_filename(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            app = Flask("src_app")
            app.config.from_object("tests.config_security_cli")
            with app.app_context():
                SQLA = get_sqla_class()
                db = SQLA(app)
                app_builder = AppBuilder(app, db.session)  # noqa: F841

                owd = os.getcwd()
                os.chdir(tmp_dir)
                cli_runner = app.test_cli_runner()
                export_result = cli_runner.invoke(export_roles)
                os.chdir(owd)

                self.assertEqual(export_result.exit_code, 0)
                self.assertGreater(
                    len(glob.glob(os.path.join(tmp_dir, "roles_export_*"))), 0
                )

    @patch("json.dumps")
    def test_export_roles_indent(self, mock_json_dumps):
        """Test that json.dumps is called with the correct argument passed from CLI."""
        app = Flask("src_app")
        app.config.from_object("tests.config_security_cli")
        with app.app_context():
            SQLA = get_sqla_class()
            db = SQLA(app)
            app_builder = AppBuilder(app, db.session)  # noqa: F841
            cli_runner = app.test_cli_runner()

            cli_runner.invoke(export_roles)
            mock_json_dumps.assert_called_with(ANY, indent=None)
            mock_json_dumps.reset_mock()

            example_cli_args = ["", "foo", -1, 0, 1]
            for arg in example_cli_args:
                cli_runner.invoke(export_roles, [f"--indent={arg}"])
                mock_json_dumps.assert_called_with(ANY, indent=arg)
                mock_json_dumps.reset_mock()

    def test_import_roles(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            app = Flask("dst_app")
            app.config[
                "SQLALCHEMY_DATABASE_URI"
            ] = f"sqlite:///{os.path.join(tmp_dir, 'dst.db')}"
            with app.app_context():
                SQLA = get_sqla_class()
                db = SQLA(app)
                app_builder = AppBuilder(app, db.session)
                cli_runner = app.test_cli_runner()

                path = os.path.join(tmp_dir, "roles.json")

                with open(path, "w") as fd:
                    fd.write(json.dumps(self.expected_roles))

                # before import roles on dst app include only Admin and Public
                self.assertEqual(len(app_builder.sm.get_all_roles()), 2)

                import_result = cli_runner.invoke(import_roles, [f"--path={path}"])
                self.assertEqual(import_result.exit_code, 0)

                resulting_roles = app_builder.sm.get_all_roles()

                for expected_role in self.expected_roles:
                    match = [
                        r for r in resulting_roles if r.name == expected_role["name"]
                    ]
                    self.assertTrue(match)
                    resulting_role = match[0]

                    expected_role_permission_view_menus = {
                        (pvm["permission"]["name"], pvm["view_menu"]["name"])
                        for pvm in expected_role["permissions"]
                    }
                    resulting_role_permission_view_menus = {
                        (pvm.permission.name, pvm.view_menu.name)
                        for pvm in resulting_role.permissions
                    }
                    self.assertEqual(
                        resulting_role_permission_view_menus,
                        expected_role_permission_view_menus,
                    )



from typing import Any, Callable, Dict, List, Optional, Type

from flask import current_app
from flask_appbuilder.models.sqla import Model
from flask_appbuilder.models.sqla.interface import SQLAInterface
from marshmallow import fields, Schema
from marshmallow.fields import Field
from marshmallow_sqlalchemy import field_for
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema


class TreeNode:
    def __init__(self, name: str) -> None:
        self.name = name
        self.children: List["TreeNode"] = []

    def __repr__(self) -> str:
        return f"{self.name}.{str(self.children)}"


class Tree:
    """
    Simplistic one level Tree
    """

    def __init__(self) -> None:
        self.root = TreeNode("+")

    def add(self, name: str) -> None:
        node = TreeNode(name)
        self.root.children.append(node)

    def add_child(self, parent: str, name: str) -> None:
        node = TreeNode(name)
        for child in self.root.children:
            if child.name == parent:
                child.children.append(node)
                return
        root = TreeNode(parent)
        self.root.children.append(root)
        root.children.append(node)

    def __repr__(self) -> str:
        ret = ""
        for node in self.root.children:
            ret += str(node)
        return ret


def columns2Tree(columns: List[str]) -> Tree:
    tree = Tree()
    for column in columns:
        if "." in column:
            parent, child = column.split(".")
            tree.add_child(parent, child)
        else:
            tree.add(column)
    return tree


class BaseModel2SchemaConverter(object):
    def __init__(
        self,
        datamodel: SQLAInterface,
        validators_columns: Dict[str, Callable[[Any], Any]],
    ):
        """
        :param datamodel: SQLAInterface
        """
        self.datamodel = datamodel
        self.validators_columns = validators_columns

    def convert(
        self,
        columns: List[str],
        model: Optional[Type[Model]] = None,
        nested: bool = True,
        parent_schema_name: Optional[str] = None,
    ) -> SQLAlchemyAutoSchema:
        pass


class Model2SchemaConverter(BaseModel2SchemaConverter):
    """
    Class that converts Models to marshmallow Schemas
    """

    def __init__(
        self,
        datamodel: SQLAInterface,
        validators_columns: Dict[str, Callable[[Any], Any]],
    ):
        """
        :param datamodel: SQLAInterface
        """
        super(Model2SchemaConverter, self).__init__(datamodel, validators_columns)

    @staticmethod
    def _debug_schema(schema: SQLAlchemyAutoSchema) -> None:
        for k, v in schema._declared_fields.items():
            print(k, v)

    def _meta_schema_factory(
        self,
        columns: List[str],
        model: Optional[Type[Model]],
        class_mixin: Type[Schema],
        parent_schema_name: Optional[str] = None,
    ) -> Type[SQLAlchemyAutoSchema]:
        """
        Creates ModelSchema marshmallow-sqlalchemy

        :param columns: a list of columns to mix
        :param model: Model
        :param class_mixin: a marshamallow Schema to mix
        :return: ModelSchema
        """
        _model = model
        _parent_schema_name = parent_schema_name
        if columns:

            class MetaSchema(SQLAlchemyAutoSchema, class_mixin):  # type: ignore
                class Meta:
                    model = _model
                    fields = columns
                    load_instance = True
                    sqla_session = current_app.appbuilder.session
                    # The parent_schema_name is useful to humanize nested schema names
                    # This name comes from ModelRestApi
                    parent_schema_name = _parent_schema_name

            return MetaSchema

        class MetaSchema(SQLAlchemyAutoSchema, class_mixin):  # type: ignore
            class Meta:
                model = _model
                load_instance = True
                sqla_session = current_app.appbuilder.session
                # The parent_schema_name is useful to humanize nested schema names
                # This name comes from ModelRestApi
                parent_schema_name = _parent_schema_name

        return MetaSchema

    def _column2enum(self, datamodel: SQLAInterface, column: TreeNode) -> Field:
        required = not datamodel.is_nullable(column.name)
        sqla_column = datamodel.list_columns[column.name]
        # get SQLAlchemy column user info, we use it to get the marshmallow enum options
        column_info = sqla_column.info
        # TODO: Default should be False, but keeping this to True to keep compatibility
        # Turn this to False in the next major release
        by_value = column_info.get("marshmallow_by_value", True)
        # Get the original enum class from SQLAlchemy Enum field
        enum_class = sqla_column.type.enum_class
        if not enum_class:
            field = field_for(datamodel.obj, column.name)
        else:
            field = fields.Enum(enum_class, required=required, by_value=by_value)
        field.unique = datamodel.is_unique(column.name)
        return field

    def _column2relation(
        self,
        datamodel: SQLAInterface,
        column: TreeNode,
        nested: bool = False,
        parent_schema_name: Optional[str] = None,
    ) -> Field:
        if nested:
            required = not datamodel.is_nullable(column.name)
            nested_model = datamodel.get_related_model(column.name)
            lst = [item.name for item in column.children]
            nested_schema = self.convert(
                lst, nested_model, nested=False, parent_schema_name=parent_schema_name
            )
            if datamodel.is_relation_many_to_one(column.name):
                many = False
            elif datamodel.is_relation_many_to_many(column.name):
                many = True
                required = False
            elif datamodel.is_relation_one_to_many(column.name):
                many = True
            else:
                many = False
            field = fields.Nested(nested_schema, many=many, required=required)
            field.unique = datamodel.is_unique(column.name)
            return field
        # Handle bug on marshmallow-sqlalchemy
        # https://github.com/marshmallow-code/marshmallow-sqlalchemy/issues/163
        if datamodel.is_relation_many_to_many(
            column.name
        ) or datamodel.is_relation_one_to_many(column.name):
            required = datamodel.get_info(column.name).get("required", False)
        else:
            required = not datamodel.is_nullable(column.name)
        field = field_for(datamodel.obj, column.name)
        field.required = required
        field.unique = datamodel.is_unique(column.name)
        return field

    def _column2field(
        self,
        datamodel: SQLAInterface,
        column: TreeNode,
        nested: bool = True,
        parent_schema_name: Optional[str] = None,
    ) -> Field:
        """

        :param datamodel: SQLAInterface
        :param column: TreeNode column (childs are dotted columns)
        :param nested: Boolean if will create nested fields
        :return: Schema.field
        """
        # Handle relations
        if datamodel.is_relation(column.name):
            return self._column2relation(
                datamodel, column, nested=nested, parent_schema_name=parent_schema_name
            )
        # Handle Enums
        if datamodel.is_enum(column.name):
            return self._column2enum(datamodel, column)
        # is custom property method field?
        if hasattr(getattr(datamodel.obj, column.name), "fget"):
            return fields.Raw(dump_only=True)
        # its a model function
        if hasattr(getattr(datamodel.obj, column.name), "__call__"):
            return fields.Function(getattr(datamodel.obj, column.name), dump_only=True)
        # is a normal model field not a function?
        if not hasattr(getattr(datamodel.obj, column.name), "__call__"):
            field = field_for(datamodel.obj, column.name)
            field.unique = datamodel.is_unique(column.name)
            if column.name in self.validators_columns:
                if field.validate is None:
                    field.validate = []
                field.validate.append(self.validators_columns[column.name])
                field.validators.append(self.validators_columns[column.name])
            return field

    def convert(
        self,
        columns: List[str],
        model: Optional[Type[Model]] = None,
        nested: bool = True,
        parent_schema_name: Optional[str] = None,
    ) -> SQLAlchemyAutoSchema:
        """
            Creates a Marshmallow ModelSchema class


        :param columns: List with columns to include, if empty converts all on model
        :param model: Override Model to convert
        :param nested: Generate relation with nested schemas
        :return: ModelSchema object
        """
        super(Model2SchemaConverter, self).convert(
            columns, model=model, nested=nested, parent_schema_name=parent_schema_name
        )

        class SchemaMixin:
            pass

        _model = model or self.datamodel.obj
        _datamodel = self.datamodel.__class__(_model)

        ma_sqla_fields_override = {}

        _columns = list()
        tree_columns = columns2Tree(columns)
        for column in tree_columns.root.children:
            # Get child model is column is dotted notation
            ma_sqla_fields_override[column.name] = self._column2field(
                _datamodel, column, nested, parent_schema_name=parent_schema_name
            )
            _columns.append(column.name)
        for k, v in ma_sqla_fields_override.items():
            setattr(SchemaMixin, k, v)
        return self._meta_schema_factory(
            _columns, _model, SchemaMixin, parent_schema_name=parent_schema_name
        )()



from app import db
from app.models import ContactGroup

db.create_all()

try:
    db.session.add(ContactGroup(name="Friends"))
    db.session.add(ContactGroup(name="Family"))
    db.session.add(ContactGroup(name="Work"))
    db.session.commit()
except:
    db.session.rollback()



import logging

from flask import Flask
from flask_appbuilder import AppBuilder, BaseView
from flask_appbuilder.api import BaseApi, expose
from flask_appbuilder.security.decorators import limit
from flask_appbuilder.utils.legacy import get_sqla_class
import hiro
import jinja2
from tests.base import FABTestCase
from tests.const import INVALID_LOGIN_STRING, PASSWORD_ADMIN, USERNAME_ADMIN


class LimiterTestCase(FABTestCase):
    def setUp(self):
        self.app = Flask(__name__)
        self.app.jinja_env.undefined = jinja2.StrictUndefined
        self.app.config.from_object("tests.config_api")
        self.app.config["RATELIMIT_ENABLED"] = True
        self.app.config["AUTH_RATE_LIMITED"] = True
        self.app.config["AUTH_RATE_LIMIT"] = "2 per 5 second"
        logging.basicConfig(level=logging.ERROR)

        self.ctx = self.app.app_context()
        self.ctx.push()

        SQLA = get_sqla_class()
        self.db = SQLA(self.app)
        self.appbuilder = AppBuilder(self.app, self.db.session)
        self.create_default_users(self.appbuilder)

        class Base1Api(BaseApi):
            @limit("2 per second")
            @expose("/test1")
            def test1(self, **kwargs):
                return self.response(200, message="OK")

        class TestView(BaseView):
            @limit("1 per second")
            @expose("/test")
            def test(self):
                return self.render_template(template=self.appbuilder.base_template)

        self.appbuilder.add_api(Base1Api)
        self.appbuilder.add_view(TestView, name="testview")

    def test_default_auth_rate_limit(self):
        client = self.app.test_client()

        with hiro.Timeline().freeze():
            rv = self.browser_login(client, USERNAME_ADMIN, "wrong_password")
            data = rv.data.decode("utf-8")
            self.assertIn(INVALID_LOGIN_STRING, data)

            rv = self.browser_login(client, USERNAME_ADMIN, "wrong_password")
            data = rv.data.decode("utf-8")
            self.assertIn(INVALID_LOGIN_STRING, data)

            rv = self.browser_login(client, USERNAME_ADMIN, "wrong_password")
            self.assertEqual(rv.status_code, 429)

    def test_api_rate_decorated_limit(self):
        client = self.app.test_client()
        token = self.login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        uri = "api/v1/base1api/test1"
        with hiro.Timeline().freeze():
            rv = self.auth_client_get(client, token, uri)
            self.assertEqual(rv.status_code, 200)

            rv = self.auth_client_get(client, token, uri)
            self.assertEqual(rv.status_code, 200)

            rv = self.auth_client_get(client, token, uri)
            self.assertEqual(rv.status_code, 429)

    def view_rate_decorated_limit(self):
        client = self.app.test_client()
        token = self.login(client, USERNAME_ADMIN, PASSWORD_ADMIN)
        uri = "/test"
        with hiro.Timeline().freeze():
            rv = self.auth_client_get(client, token, uri)
            self.assertEqual(rv.status_code, 200)

            rv = self.auth_client_get(client, token, uri)
            self.assertEqual(rv.status_code, 429)



import logging

from flask import Flask
from sqlalchemy import event
from sqlalchemy.engine import Engine
from .extensions import appbuilder, db

logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def create_app() -> Flask:
    app = Flask(__name__)
    app.config.from_object("config")
    with app.app_context():
        db.init_app(app)
        appbuilder.init_app(app, db.session)
        from . import views, models  # noqa
    return app


app = create_app()



from flask_appbuilder import ModelRestApi
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder.security.sqla.models import PermissionView


class PermissionViewMenuApi(ModelRestApi):
    resource_name = "security/permissions-resources"
    openapi_spec_tag = "Security Permissions on Resources (View Menus)"
    class_permission_name = "PermissionViewMenu"
    datamodel = SQLAInterface(PermissionView)
    allow_browser_login = True

    list_columns = ["id", "permission.name", "view_menu.name"]
    show_columns = list_columns
    add_columns = ["permission_id", "view_menu_id"]
    edit_columns = add_columns



import logging
from app import create_app
from app.models import Datacenter, Rack, Item
from flask_appbuilder.extensions import db
import random
import string

log = logging.getLogger(__name__)

DC_RACK_MAX = 20
ITEM_MAX = 1000

cities = ["Lisbon", "Porto", "Madrid", "Barcelona", "Frankfurt", "London"]

models = ["Server MX", "Server MY", "Server DL380", "Server x440", "Server x460"]

datacenters = list()

app = create_app()


def get_random_name(names_list, size=1):
    return names_list[random.randrange(0, len(names_list))]


def serial_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return "".join(random.choice(chars) for _ in range(size))


def upsert_test_data():
    for city in cities:
        datacenter = Datacenter()
        datacenter.name = "DC %s" % city
        datacenter.address = city
        datacenters.append(datacenter)
        db.session.add(datacenter)
        log.info(datacenter)
        try:
            db.session.commit()
            for num in range(1, DC_RACK_MAX):
                rack = Rack()
                rack.num = num
                rack.datacenter = datacenter
                db.session.add(rack)

        except Exception as e:
            log.error("Creating Datacenter: %s", e)
            db.session.rollback()

    for i in range(1, ITEM_MAX):
        item = Item()
        item.serial_number = serial_generator()
        item.model = get_random_name(models)
        db.session.add(item)
        log.info(item)
        try:
            db.session.commit()
        except Exception as e:
            log.error("Creating Item: %s", e)
            db.session.rollback()


with app.app_context():
    upsert_test_data()
